/*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
// (C) PicoPeta Simputers Pvt. Ltd. (2003)
// Author: Vivek K S

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include "alchemy.h"
#include "ewmh.h"
#include "normal_window.h"
#include "basic_window.h"

// A new normal window
window *new_normal_window(wm *wm, Window map_window)
{
	window *win;
	XSetWindowAttributes attr;
	unsigned int topbar_height, bottombar_height;
	ac_theme *theme = wm->current_theme;

	// Allocate and assign common data structures
	win=new_basic_window(wm, map_window, NORMAL_WIN);

	// Assign unallocated pointers to NULL.
	// Set Function pointers
	win->redraw = &normal_window_redraw;
	win->refresh_widgets = &normal_window_refresh_widgets;
	win->set_size = &normal_window_set_size;
	win->set_position = &normal_window_set_position;
	win->show = &normal_window_show;
	win->hide = &normal_window_hide;
	win->iconify = &normal_window_hide;
	win->destroy = &normal_window_destroy;
	win->set_flags = &normal_window_set_flags;
	win->set_flag = &normal_window_set_flag;
	win->app_window = map_window;

	// Figure out top and bottombar sizes
	topbar_height = theme->get_topbar_height(theme);
	bottombar_height = theme->get_bottombar_height(theme);

	// Set sizes and position
	win->x = 0;
	win->width = wm->display_width;

	win->y = topbar_height;
	win->height = wm->display_height - (wm->get_toolbar_displacement(wm) + topbar_height + bottombar_height);

	// Reparenting happens here
	// Create toplevel, topbar and bottombar frames
	attr.override_redirect = True;
	attr.border_pixel = attr.background_pixel = wm->black_pixel;
	attr.event_mask = SubstructureRedirectMask|SubstructureNotifyMask|
		ButtonPressMask|ButtonReleaseMask|EnterWindowMask|
		LeaveWindowMask|ExposureMask;

	win->frame_window = XCreateWindow(wm->display, wm->root, 0, 0, 
			win->width, win->height + topbar_height
			+ bottombar_height, 0,
			CopyFromParent, CopyFromParent, CopyFromParent,
			CWOverrideRedirect|CWBorderPixel|CWEventMask|CWBackPixel,
			&attr);

	// Render the theme
	theme->render_theme(theme, win);

	// Make cursors as either null (for simputer) or arrows for PC
	XDefineCursor(wm->display, win->topbar_frame->win, wm->normal_cursor);
	XDefineCursor(wm->display, win->bottombar_frame->win, wm->normal_cursor);

	XSelectInput(wm->display, win->app_window,
		ColormapChangeMask|PropertyChangeMask);

	XReparentWindow(wm->display, win->app_window, win->frame_window,
		0, topbar_height);

	XResizeWindow(wm->display, win->app_window, win->width, win->height);
	//window_send_configure_event(win); // seems unneccessary
	XSetWindowBorderWidth(wm->display, win->app_window, 0);

	// This ensures that the windows are not killed even when we exit
	XAddToSaveSet(wm->display, win->app_window);

	// Handle full screen
	if(win->flags.bits.full_screen)
		normal_window_handle_full_screen(win);

	return win;
}

// Set the size of the window, resize frame to accomodate the app_window
// send configure event, re-position the bottom bar
void normal_window_set_size(window *win, unsigned long width, unsigned long height)
{
	int topbar_height, bottombar_height;
	wm* wm=win->wm;
	ac_theme *theme = wm->current_theme;

	win->width=width;
	win->height=height;

	XResizeWindow(wm->display, win->app_window, width, height);
	if(win->flags.bits.full_screen) {
		XResizeWindow(wm->display, win->frame_window, wm->display_width, wm->display_height);
	} else {
		topbar_height = theme->get_topbar_height(theme);
		bottombar_height = theme->get_bottombar_height(theme);
		XResizeWindow(wm->display, win->frame_window, width, height + topbar_height + bottombar_height);

		win->bottombar_frame->set_position(win->bottombar_frame, 0, 
			win->y + win->height);
	}
	// Seems unecessary
	window_send_configure_event(win);
}

// Set the position of the frame window, app window automatically moves
void normal_window_set_position(window *win, int x, int y)
{
	Display *dpy = win->wm->display;
	int topbar_height = win->wm->current_theme->get_topbar_height(win->wm->current_theme);

	win->x=x;
	win->y=y;
	if(win->flags.bits.full_screen) {
		XMoveWindow(dpy, win->app_window, x, y );
		XMoveWindow(dpy, win->frame_window, 0, 0 );
	} else {
		XMoveWindow(dpy, win->app_window, 0, topbar_height);
		XMoveWindow(dpy, win->frame_window, x, y - topbar_height);
	}
}

// Shift from fullscreen mode to normal or vice-versa
void normal_window_handle_full_screen(window *win)
{
	wm *wm = win->wm;
	ac_theme *theme = wm->current_theme;
	int topbar_height, bottombar_height;

	topbar_height = theme->get_topbar_height(theme);
	bottombar_height = theme->get_bottombar_height(theme);

	if(win->flags.bits.full_screen) {
		// hide keyboard if mapped
		if(wm->keyboard_window && wm->is_keyboard_mapped)
			wm->keyboard_window->hide(wm->keyboard_window);

		win->set_size(win, wm->display_width, wm->display_height);
		win->bottombar_frame->set_position(win->bottombar_frame, 0, -bottombar_height);
		win->topbar_frame->set_position(win->topbar_frame, 0, -topbar_height);
		win->set_position(win, 0, 0);
	} else {
		win->set_position(win, win->x, topbar_height);
		win->set_size(win, wm->display_width, 
				wm->display_height - (topbar_height + bottombar_height));
		win->topbar_frame->set_position(win->topbar_frame, 0, 0);
		win->bottombar_frame->set_position(win->bottombar_frame, 0, wm->display_height - bottombar_height);
		window_handle_keyboard(win);
	}
}

// Redraw the top and bottom bars
void normal_window_redraw(window *win)
{
	win->wm->current_theme->render_theme(win->wm->current_theme, win);
}

// Used during dynamic theme changing
void normal_window_refresh_widgets(window *win)
{
	win->topbar_frame->redraw(win->topbar_frame);
	win->bottombar_frame->redraw(win->bottombar_frame);
	win->topbar_frame->show(win->topbar_frame);
	win->bottombar_frame->show(win->bottombar_frame);
}

// Lower this window and unmap its transients
void normal_window_hide(window *win)
{
	if(win->transient_next)
		win->transient_next->hide(win->transient_next);

	window_set_state(win, IconicState);
	win->is_mapped = False;
	XLowerWindow(win->wm->display, win->frame_window);
}

// Bring up this window along with the transients
void normal_window_show(window *win)
{
	winlist *winlist_node;
	wm *wm = win->wm;

	if(win->layer_above)
		return;

	if(win->fast_load && !win->fast_visible)
		return;

	window_set_state(win, NormalState);
	
	XMapRaised(wm->display, win->frame_window);
	XMapSubwindows(wm->display, win->frame_window);

	// Raise the top and bottom frames... so that
	// no app can stay on top of this
	win->topbar_frame->raise(win->topbar_frame);
	win->topbar_frame->raise(win->topbar_frame);

	wm->mapped_window = win;
	win->is_mapped = True;

#ifdef __arm__
	window_handle_rotation(win);	// Rotate display if necessary
#endif
	window_handle_keyboard(win);	// Show/hide keyboard if required
	window_handle_annotate_icon(win);	// Show correct annotation icon
	xstroke_stop_recognition(wm);	// Stop xstroke

	if(win->transient_next)
		win->transient_next->show(win->transient_next);

	window_handle_focus(win);

	ac_ewmh_set_active_window(wm);

	if(wm->clock_window && win->appid == AC_LAUNCHER_APPID && !win->layer_below) {
		wm->clock_window->show(wm->clock_window);
	}

	// Vivek: new, map all the on_top list windows
	winlist_node = wm->on_top_list;
	while(winlist_node) {
		winlist_node->win->show(winlist_node->win);
		winlist_node=winlist_node->next;
	}
}

// Kill this window
void normal_window_destroy(window *win)
{
	window *t;
	wm *wm = win->wm;

	if(!win->layer_above) {
		if(win->appid == AC_IDENTITY_APPID &&
			window_exists(wm, wm->identity_save_window)) {
			wm->identity_save_window->show(wm->identity_save_window);
		} else if(win->layer_below) {
			window *sw = win->layer_below;

			sw->layer_above = NULL;	// Hack to make sure that
						// show is called
			sw->show(sw);
			sw->layer_above = win;	// Restore link
		} else if(win->launcher_appid != AC_LAUNCHER_APPID 
			&& ((t=window_find_appid(wm, win->launcher_appid))!=NULL) ) {
			t->show(t);
		} else if(wm->launcher_window && win != wm->launcher_window ) {
			window *launch_top=wm->launcher_window;

			/* Navigate through the layers till the top */
			while(launch_top->layer_above)
				launch_top=launch_top->layer_above;
			launch_top->show(launch_top);
		} else if((t=window_get_next_mappable(win)) != win) {
			t->show(t);
		} else if(wm->window_list != win) {
			wm->window_list->show(wm->window_list);
		}
	}

	t=win->transient_next;

	while(t) {
		window *t2;

		t2=t->transient_next;		// Save the net pointer before it disappears
		window_handle_deletion(t);	// delete t
		t=t2;						// Move to next window
	}

	window_destroy(win);
}

// Set the flags and atom
void normal_window_set_flags(window *win, unsigned long flags)
{
	if(win->appid == AC_LAUNCHER_APPID && !win->layer_below)
		return;

	win->flags.all_flags = flags;
	window_set_int_property(win, _AC_W_FLAGS, flags );
}

// Set the flag and atom
void normal_window_set_flag(window *win, flag_num flag_no, Bool value)
{
	if(win->appid == AC_LAUNCHER_APPID && !win->layer_below)
		return;

	if(value)
		win->flags.all_flags |= (1<<flag_no);
	else
		win->flags.all_flags &= (~(1<<flag_no));

	window_set_int_property(win, _AC_W_FLAGS, win->flags.all_flags );
}
