/*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
// (C) PicoPeta Simputers Pvt. Ltd. (2003)
// Author: Vivek K S

#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>
#include <unistd.h>
#include <X11/Xlib.h>
#include <errno.h>
#include "utils.h"

unsigned char xerror_variable; // To detect errors handled by xerror_handler

// Print an error message on stderr and exit
void err_exit(const char *fmt, ...)
{
    va_list argp;
    fprintf(stderr, "Error: ");
    va_start(argp, fmt);
    vfprintf(stderr, fmt, argp);
    va_end(argp);
    fprintf(stderr, "\n");
	exit(1);
} 

// Print some debug info on stderr
void debug_info(const char *fmt, ...)
{
    va_list argp;
    fprintf(stderr, "Debug: ");
    va_start(argp, fmt);
    vfprintf(stderr, fmt, argp);
    va_end(argp);
    fprintf(stderr, "\n");
} 

// The X Error handler.. invoked for every X Error that occurs
int xerror_handler(Display *display, XErrorEvent *xev)
{
	char error_msg[512];

	if(xev->error_code == BadAccess && 
			xev->resourceid == RootWindow(display, DefaultScreen(display))) {
		err_exit("Unable to manage root window (Check if another wm is running.)");
	}
	xerror_variable = xev->error_code;

	XGetErrorText(display, xev->error_code, error_msg, sizeof (error_msg));
	debug_info("X Error (#%lx): %s for request %lx", xev->resourceid, error_msg, xev->request_code);
	return 0;
}

static void sigchld_handler(int signo)
{
   int status;

    /* Wait for any child without blocking */
    if (waitpid(-1, &status, WNOHANG) < 0)
        debug_info("waitpid failed");
}

void init_signals()
{
	struct sigaction act;

	/* Assign sigchld_handler as the SIGCHLD handler */
	act.sa_handler = sigchld_handler;

	sigemptyset(&act.sa_mask);

	/*
	* We're only interested in children that have terminated, not ones
	* which have been stopped
	*/
	act.sa_flags = SA_NOCLDSTOP;

	if (sigaction(SIGCHLD, &act, NULL) < 0) 
		err_exit("sigaction failed");
}

// Spawn a command, pass it some environment variables
void spawn_child(const char *cmd, ...)
{
	va_list envlist;
	pid_t pid=fork();

	switch(pid) {
		char *envp;

		case 0:
			setsid();
			va_start(envlist, cmd);
			while((envp = va_arg(envlist, char *)) != NULL)
				if(putenv(envp))
					debug_info("Could not putenv: %s", envp);
				else {
					DBG(puts(envp);)
				}
			va_end(envlist);
#ifdef __arm__
			putenv("DISPLAY=:0.0");
			putenv("SHELL=/bin/sh");
			putenv("LD_LIBRARY_PATH=/lib:/usr/lib:/usr/local/lib");
			putenv("PATH=/bin:/usr/bin:/usr/local/bin:/sbin:/usr/sbin");
#endif
			execlp("/bin/sh","/bin/sh","-c",cmd,NULL);
			perror("Could not execlp");
			debug_info("Could not execlp : %s",cmd);
			exit(1);
			break;

		case -1:
			debug_info("Could not fork a child for: %s",cmd);
			break;

		default:
			break;
	}
}
